<?php

namespace EventPortal\Tests\Unit\ActionList;

use EventPortal\Domain\ActionList\Models\ActionList;
use EventPortal\Domain\ActionList\Models\Task as TaskModel;
use EventPortal\Domain\ActionList\Task;
use EventPortal\Domain\Admin\Models\Admin;
use EventPortal\Domain\Event\Models\Event;
use EventPortal\Tests\Support\Traits\SeedRolesAndPermissions;
use EventPortal\Tests\TestCase;
use Illuminate\Foundation\Testing\RefreshDatabase;

class TaskTest extends TestCase
{
    use RefreshDatabase, SeedRolesAndPermissions;

    /** @test **/
    public function can_add_label()
    {
        $task = Task::make('name_of_task')->label('This is the label');

        $this->assertSame('name_of_task', $task->name);
        $this->assertSame('This is the label', $task->label);
    }

    /** @test **/
    public function not_passing_a_name_converts_label_to_name()
    {
        $task = Task::make()->label('This is the label');

        $this->assertSame('this_is_the_label', $task->name);
    }

    /** @test **/
    public function can_mark_as_required()
    {
        $task = Task::make()->label('The task');

        $this->assertFalse($task->required);

        $task->required();

        $this->assertTrue($task->required);
    }

    /** @test **/
    public function can_mark_as_automatic()
    {
        $task = Task::make()->label('The task');

        $this->assertFalse($task->automatic);

        $task->automatic();

        $this->assertTrue($task->automatic);
    }

    /** @test **/
    public function can_add_href()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->href);

        $task->href('https://google.com');

        $this->assertSame('https://google.com', $task->href);
    }

    /** @test **/
    public function can_add_route()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->route);
        $this->assertNull($task->routeParams);
        $this->assertNull($task->actionLabel);

        $task->route('test.route');

        $this->assertSame('test.route', $task->route);
        $this->assertSame([], $task->routeParams);
        $this->assertNull($task->actionLabel);
    }

    /** @test **/
    public function can_add_route_params()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->route);
        $this->assertNull($task->routeParams);
        $this->assertNull($task->actionLabel);

        $task->route('test.route', ['id' => 1]);

        $this->assertSame('test.route', $task->route);
        $this->assertSame(['id' => 1], $task->routeParams);
        $this->assertNull($task->actionLabel);
    }

    /** @test **/
    public function can_add_route_action_label()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->route);
        $this->assertNull($task->routeParams);
        $this->assertNull($task->actionLabel);

        $task->route('test.route', ['id' => 1], 'Do Something');

        $this->assertSame('test.route', $task->route);
        $this->assertSame(['id' => 1], $task->routeParams);
        $this->assertSame('Do Something', $task->actionLabel);
    }

    /** @test **/
    public function can_add_action_label()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->actionLabel);

        $task->actionLabel('Do Something');

        $this->assertSame('Do Something', $task->actionLabel);
    }

    /** @test **/
    public function can_add_event()
    {
        $event = factory(Event::class)->create();
        $task = Task::make()->label('The task');

        $this->assertNull($task->event);

        $task->event($event);

        $this->assertSameModel($event, $task->event);
    }

    /** @test **/
    public function can_get_action_for_href()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->getAction());

        $task->href('https://google.com', 'Go to Google');

        $this->assertSame(
            ['type' => 'href', 'route' => 'https://google.com', 'label' => 'Go to Google'],
            $task->getAction()
        );
    }

    /** @test **/
    public function can_get_action_for_route()
    {
        $task = Task::make()->label('The task');

        $this->assertNull($task->getAction());

        $task->route('test.route');
        $this->assertSame(
            ['type' => 'route', 'route' => 'test.route', 'routeParams' => [], 'label' => null],
            $task->getAction()
        );

        $task->route('test.route', ['id' => 1]);
        $this->assertSame(
            ['type' => 'route', 'route' => 'test.route', 'routeParams' => ['id' => 1], 'label' => null],
            $task->getAction()
        );

        $task->route('test.route', ['id' => 1], 'Go to route');
        $this->assertSame(
            ['type' => 'route', 'route' => 'test.route', 'routeParams' => ['id' => 1], 'label' => 'Go to route'],
            $task->getAction()
        );
    }

    /** @test **/
    public function get_returns_empty_array_if_no_event()
    {
        $task = Task::make()->label('The task');

        $this->assertSame([], $task->get());
    }

    /** @test **/
    public function get_returns_empty_array_if_no_action_list()
    {
        $event = factory(Event::class)->create();
        $task = Task::make()->label('The task');

        $task->event($event);

        $this->assertSame([], $task->get());
    }

    /** @test **/
    public function get_returns_tasks()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => null,
                'completed' => false,
                'automatic' => false,
              ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test **/
    public function get_returns_tasks_with_href()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
            'action' => '{"type":"href","route":"https://google.com","label":"Go to Google"}',
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => '{"type":"href","route":"https://google.com","label":"Go to Google"}',
                'completed' => false,
                'automatic' => false,
            ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test **/
    public function get_returns_tasks_with_route()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
            'action' => '{"type":"route","route":"test.route","routeParams":{"id":1},"label":"Go to route"}',
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => '{"type":"route","route":"test.route","routeParams":{"id":1},"label":"Go to route"}',
                'completed' => false,
                'automatic' => false,
            ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test **/
    public function get_returns_tasks_that_have_been_completed()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
            'completed_at' => '2021-02-32 12:44:00',
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => null,
                'completed' => true,
                'automatic' => false,
            ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test **/
    public function get_returns_tasks_that_is_automatic()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
            'automatic' => true,
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => null,
                'completed' => false,
                'automatic' => true,
            ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test **/
    public function get_returns_multiple_tasks()
    {
        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task',
            'label' => 'The task',
        ]);

        factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'the_task_2',
            'label' => 'The task 2',
        ]);

        $task = Task::make('the_task')
            ->label('The task')
            ->event($event);

        $expected = [
            [
                'id' => 1,
                'label' => 'The task',
                'action' => null,
                'completed' => false,
                'automatic' => false,
            ],
            [
                'id' => 2,
                'label' => 'The task 2',
                'action' => null,
                'completed' => false,
                'automatic' => false,
            ],
        ];

        $this->assertSame($expected, $task->get());
    }

    /** @test */
    public function complete_a_task_that_exists()
    {
        $admin = factory(Admin::class)->create();

        $this->actingAsAdmin($admin);

        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);

        $task = factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'venue',
        ]);

        $this->assertNull($task->completed_at);

        Task::make('venue')->event($event)->complete($admin->user);

        $this->assertEquals(now(), $task->fresh()->completed_at);
        $this->assertEquals(1, $task->fresh()->user_id);
    }

    /** @test */
    public function marks_complete_for_correct_event()
    {
        $admin = factory(Admin::class)->create();

        $this->actingAsAdmin($admin);

        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);
        $actionList = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event->id,
        ]);
        $task = factory(TaskModel::class)->create([
            'action_list_id' => $actionList->id,
            'name' => 'venue',
        ]);

        $event2 = factory(Event::class)->create([
            'status' => 'test',
        ]);
        $actionList2 = factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
            'event_id' => $event2->id,
        ]);
        $task2 = factory(TaskModel::class)->create([
            'action_list_id' => $actionList2->id,
            'name' => 'venue',
        ]);

        $this->assertNull($task->completed_at);
        $this->assertNull($task2->completed_at);

        Task::make('venue')->event($event2)->complete($admin->user);

        $this->assertNull($task->fresh()->completed_at);
        $this->assertNull($task->fresh()->user_id);

        $this->assertEquals(now(), $task2->fresh()->completed_at);
        $this->assertEquals(1, $task2->fresh()->user_id);
    }

    /** @test */
    public function does_not_error_if_action_list_does_not_exist()
    {
        $admin = factory(Admin::class)->create();

        $this->actingAsAdmin($admin);

        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        try {
            Task::make('does not exist')->event($event)->complete($admin->user);
        } catch (\Throwable $exception) {
            $this->fail('Should not have thrown an error. Should be able to handle an action list not existing');
            return;
        }

        $this->addToAssertionCount(1);
    }

    /** @test */
    public function does_not_error_if_task_does_not_exist()
    {
        $admin = factory(Admin::class)->create();

        $this->actingAsAdmin($admin);

        $event = factory(Event::class)->create([
            'status' => 'test',
        ]);

        factory(ActionList::class)->create([
            'status' => 'test',
            'type' => 'default',
        ]);

        try {
            Task::make('does not exist')->event($event)->complete($admin->user);
        } catch (\Throwable $exception) {
            $this->fail('Should not have thrown an error. Should be able to handle an action list not existing');
            return;
        }

        $this->addToAssertionCount(1);
    }
}
